"""Athena Query Utilities"""


def encode_list_as_list_of_strings(items: list):
    """
    Encodes a list as a String list of string items for Athena query.

    Adding a list of items to athena query argument requires it to be
    encoded as a string encoded list
    """
    items = ["'{ITEM}'".format(ITEM=item) for item in items]
    return ",".join(items)



"""
    distance scoring methods

    methods include
    ------------------------------------
    1. adjusted_cosine
    2. score_method_handler
"""
from scipy import spatial
import numpy as np


def adjusted_cosine(base_vector, comp_vector):
    """finds indices in entity_vector that equal, removes in both vectors and applies cosine

    Args:
        base_vector (list): base vector
        comp_vector (list): comparison vector

    Returns:
        float: adjusted cosine score
    """
    if isinstance(base_vector, np.ndarray):
        base_vector = base_vector.tolist()

    if isinstance(comp_vector, np.ndarray):
        comp_vector = comp_vector.tolist()

    indices = [
        i for i, x in enumerate(base_vector) if x == 0
    ]  # must be from joblisting

    indices = sorted(indices, reverse=True)

    for idx in indices:
        comp_vector.pop(idx)

    for idx in indices:
        base_vector.pop(idx)

    score = 1 - spatial.distance.cosine(base_vector, comp_vector)

    return float(score)



"""Handler classes which can be used to combined logic for scorers with similar requirements"""
from scipy import spatial
from dp_tms.scorers import distance as custom_distance_scorers


def vector_scorer_handler(
    base_vector: list, comp_vector: list, scorer_types: list
) -> dict:
    """scorer method handler for vector based scorers

    Args:
        base_vector (list): base vector
        comp_vector (list): comparison vector
        scorer_types (list): scorer types to run scoring for

    Raises:
        TypeError: _description_

    Returns:
        dict: scores for given vectors
    """
    scores = {}

    if "simple_cosine" in scorer_types:

        score = 1 - spatial.distance.cosine(base_vector, comp_vector)
        scores["simple_cosine"] = float(score)

    if "adjusted_cosine" in scorer_types:

        scores["adjusted_cosine"] = custom_distance_scorers.adjusted_cosine(
            base_vector, comp_vector
        )

    if "simple_euclidean" in scorer_types:

        # score = np.linalg.norm(np.asarray(base_1)-np.asarray(comp_1))#this is theoratically faster function
        score = 1 / (
            1 + spatial.distance.euclidean(base_vector, comp_vector)
        )  # this is a slower function, should actually use np sqrt
        scores["simple_euclidean"] = float(score)

    if len(scores.keys()) == 0:
        raise TypeError(
            f"None of `scorer_types` == {scorer_types} are currently supported"
        )

    return scores



"""query utilities and reusable methods"""
from typing import List, Dict
import numpy as np
import pandas as pd
import awswrangler as aw
from dp_tms.database.queries import entities as entity_queries


def query_from_candidate_matching_database(query: str, workgroup: str) -> pd.DataFrame:
    """run query against selected database

    Args:
        query (str): SQL query string
        workgroup (str): name of athena workgroup

    Returns:
        pd.DataFrame: results of query
    """
    dataframe = aw.athena.read_sql_query(
        query, database="dp_candidate_matching_tf", workgroup=workgroup
    )
    return dataframe


def get_vectors_by_entity_type(
    entities: list, client_code: str, entity_type: str, workgroup: str
) -> np.ndarray:
    """get entity vectors by entity type

    Args:
        entities (list): entity ids to retrieve
        client_code (str): client identifier code
        entity_type (str): type of entity to retrieve
        workgroup (str): name of athena workgroup

    Raises:
        TypeError: for unsupported entity type

    Returns:
        np.ndarray: _description_
    """
    query = entity_queries.get_jarvis_vector_embeddings(
        entities=entities, entity_type=entity_type, client_code=client_code
    )
    data = query_from_candidate_matching_database(query, workgroup)

    # order rows by entity_ids
    # and return vector
    return data.set_index("entity_id").loc[entities].vector.to_numpy()


def get_jarvis_request_metadata_by_entity_type(
    entities: list, entity_types: list, client_code: str, workgroup: str
) -> List[Dict[str, str]]:
    """get entity vectors for entity type & id for client and environment

    Args:
        entities (list): list of ids for each entity
        entity_types (list): list of types for each entity
        client_code (str): client identifier code
        workgroup (str): name of athena workgroup

    Returns:
        List[Dict[str, str]]: result from query as records
    """

    query = entity_queries.get_jarvis_request_metadata(
        entities=entities, entity_types=entity_types, client_code=client_code
    )

    data = query_from_candidate_matching_database(query, workgroup)

    return data.to_dict(orient="records")



"""
    Storage utility methods

    methods include
    --------------------------
    1. check_if_cloud_storage_and_extract_path
    2. write_txt_to_s3
    3. store_parquet_in_s3
"""
from urllib.parse import urlparse
import boto3
from botocore.client import Config
import pandas
import awswrangler as aw


def check_if_cloud_storage_and_extract_path(path: str, schemes: list = ["s3"]):
    """extracts bucket, path and scheme for select schemas or returns path otherwise

    Args:
        path (str): file path
        schemes (list, optional): schemas to extract from. Defaults to ['s3'].

    Returns:
        Tuple(str, str, str): bucket, filepath, scheme
    """
    parsed_url = urlparse(path)

    scheme = parsed_url.scheme
    if scheme in schemes:

        bucket = parsed_url.netloc
        s3_path = parsed_url.path.lstrip("/")
        return bucket, s3_path, scheme

    else:
        return None, path, None


def write_txt_to_s3(storage_path, text):
    """write a txt file to s3

    Args:
        storage_path (str): full path to file in s3
        text (str): text dump / blob to write as txt
    """
    print(f"writing txt to s3 {storage_path}")
    content = text.encode()
    config = Config(connect_timeout=10, retries={"max_attempts": 3})
    s3_client = boto3.client("s3", config=config)

    base_bucket, file_path, _ = check_if_cloud_storage_and_extract_path(storage_path)
    print(base_bucket, file_path)
    s3_client.put_object(Bucket=base_bucket, Key=file_path, Body=content)


def store_parquet_in_s3(body: dict):
    """Store content in glue table, as parquet in s3, and optiionally re-run glue crawler

    Args:
        body (dict): configuration with required fields

    Raises:
        Exception: '`content` must be supplied'
        Exception: `stage_path`, `glue_database` and `glue_table` must be supplied to stage data
    """
    # pylint: disable=too-many-locals, broad-except, too-many-branches
    (
        content,
        stage_bucket,
        stage_base_path,
        stage_filename_prefix,
        glue_database,
        glue_table,
        glue_crawler,
    ) = (
        body.get("content"),
        body.get("stage_bucket"),
        body.get("stage_base_path"),
        body.get("stage_filename_prefix", ""),
        body.get("glue_database"),
        body.get("glue_table"),
        body.get("glue_crawler"),
    )

    if not content:
        raise KeyError("`content` must be supplied")

    config = Config(connect_timeout=10, retries={"max_attempts": 3})

    if stage_bucket:
        if not (stage_base_path and glue_database and glue_table):
            raise KeyError(
                (
                    "`stage_base_path`, `glue_database` and `glue_table` "
                    "must be supplied for staging data"
                )
            )

        run_crawler = False

        if glue_crawler:
            # Check if the date partition already exists as we need to re-run the crawler
            # after the first file is placed in that partition
            s3_client = boto3.client("s3", config=config)
            response = s3_client.list_objects(
                Bucket=stage_bucket,
                Prefix=f"{stage_base_path}/{stage_filename_prefix}/",
            )

            if len(response.get("Contents", [])) == 0:
                run_crawler = True

        data_frame = pandas.DataFrame(content)

        aw.s3.to_parquet(
            df=data_frame,
            path=f"s3://{stage_bucket}/{stage_base_path}",
            mode="append",
            dataset=True,
            database=glue_database,
            table=glue_table,
            filename_prefix=f"{stage_filename_prefix}/",
        )

        print(
            f"Successfully stored staged data at s3://{stage_bucket}/{stage_base_path}/{stage_filename_prefix}/"
        )

        if glue_crawler and run_crawler:
            print(f"Running crawler {glue_crawler}")

            try:
                glue_client = boto3.client("glue")
                glue_client.start_crawler(Name=glue_crawler)
            except Exception as ex:
                print(f"Error while trying to start crawler {str(ex)}")



"""entity query string methods"""
from dp_tms.utils.athena import encode_list_as_list_of_strings


def get_jarvis_request_metadata(entities, entity_types, client_code):
    """
    Args:
        entities (list): entity ids to retrieve
        entity_types (list): list of types for each entity
        client_code (str): client identifier code
        environment (str): development staging environment to use i.e. prod, dev, test

    Returns:
        str: SQL query string
    """

    entity_list_strings = [
        f"{entity_type}-{entity_id}"
        for entity_id, entity_type in zip(entities, entity_types)
    ]
    entity_list_strings = encode_list_as_list_of_strings(items=entity_list_strings)

    return f"""
        WITH
        LABEL_BY_MOST_RECENT AS (
            SELECT
                *,
                rank() OVER (
                    PARTITION BY entity_id, entity_type, partition_0
                    ORDER BY created_at DESC
                ) AS rnk
                FROM "dp_candidate_matching_tf"."entities_raw"
        )

        SELECT
            entity_id,
            entity_type,
            properties.storage_reference.disk,
            properties.storage_reference.bucket,
            properties.storage_reference.file as file_path,
            properties.storage_reference.type as file_type
        FROM LABEL_BY_MOST_RECENT
        WHERE
            rnk = 1
            AND CONCAT(entity_type, '-', entity_id) in  ({entity_list_strings})
            AND partition_0 = '{client_code}'
        ;
    """


def get_jarvis_vector_embeddings(
    entities: list, entity_type: str, client_code: str
) -> str:
    """get a single listing vector for comparison

        note:
            - this could be multiple listings to speed up query
    Args:
        entity_ids (list): entity ids to retrieve
        entity_type (str): type of entity to retrieve
        client_code (str): client identifier code
        environment (str): development staging environment to use i.e. prod, dev, test

    Returns:
        str: SQL query string
    """

    entities = encode_list_as_list_of_strings(items=entities)
    return f"""
        WITH LABEL_BY_MOST_RECENT AS (
            SELECT
                *,
                rank() OVER (
                    PARTITION BY entity_id, entity_type, source, partition_0
                    ORDER BY created_at DESC
                ) AS rnk
                FROM "dp_candidate_matching_tf"."entities_enriched"
        )

        SELECT
            entity_id,
            properties.embedding.vector
        FROM LABEL_BY_MOST_RECENT
        WHERE
            rnk = 1
            AND source = 'jarvis'
            AND entity_type = '{entity_type}'
            AND entity_id in ({entities})
            AND partition_0 = '{client_code}'
        ;
    """



"""SETUP PACKAGING"""
from setuptools import setup

with open("public.md", "r") as fh:
    long_description = fh.read()

# for experiments
dev_requires = [
    "mpmath==1.1.0",
    "regex==2021.8.3",
    "requests==2.27.1",
]

base_requires = [
    'awswrangler==2.14.0',
    'requests==2.27.1',
    'scipy' # both scipy and awswrangler must either be locked to complimentary versions or leave scipy to find the appropriate version
]

setup(
    name="dp_tms",
    version="0.2.1",
    description="reusable utilities shared across tms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Ringier South Africa',
    author_email='tools@ringier.co.za',
    packages=[
        "dp_tms",
        "dp_tms/utils",
        "dp_tms/database",
        "dp_tms/database/queries",
        "dp_tms/scorers",
        "dp_tms/services",
    ],
    python_requires=">=3.6",
    install_requires=base_requires,
    extras_require={
        "develop": dev_requires
    }
)
